from easy_spider.filters.build_in import *
from easy_spider.network.request import Request
import unittest


class TestFilter(unittest.TestCase):

    def test_reg_filter(self):
        f = RegexFilter(r"http://")
        self.assertTrue(f.accept("http://www.baidu.com"))

    def assert_true(self, r, url):
        self.assertTrue(r.accept(url))

    def assert_false(self, r, url):
        self.assertFalse(r.accept(url))

    def test_and_op(self):
        r = RegexFilter("") + RegexFilter("") + RegexFilter("")
        self.assertEqual(type(r), AndChainFilter)

    def test_or_op(self):
        r = RegexFilter("") | RegexFilter("") | RegexFilter("")
        self.assertEqual(type(r), OrChainFilter)

    def test_and_filter(self):
        r = RegexFilter("http://") + RegexFilter(".*?test")
        self.assert_false(r, "http://www.baidu.com")
        self.assert_true(r, "http://www.test.com")

    def test_or_filter(self):
        r = RegexFilter("http://") | RegexFilter("https://")
        self.assert_true(r, "http://www.baidu.com")
        self.assert_true(r, "https://www.baidu.com")

    def test_not_filter(self):
        r = -RegexFilter("javascript:")
        self.assert_false(r, "javascript:")

    def test_and_not_filter(self):
        r = RegexFilter(".*") + (-RegexFilter("javascript:"))
        self.assert_true(r, "http://www.baidu.com")
        self.assert_false(r, "javascript:func(1)")

        r = RegexFilter(".*") - RegexFilter("javascript:") - RegexFilter("https://")
        self.assert_true(r, "http://www.baidu.com")
        self.assert_false(r, "https://www.baidu.com")
        self.assert_false(r, "javascript:func(1)")

    def test_html_filter(self):
        self.assert_true(html_filter, "http://www.baidu.com")
        self.assert_true(static_filter, "http://www.baidu.com/a.zip")
        self.assert_false(html_filter, "http://www.baidu.com/a.zip")
        self.assert_false(html_filter, "http://www.baidu.com/b.Mp4")
        self.assert_false(html_filter, "http://www.baidu.com/b.mP3")

    def test_bloom_filter(self):
        history_filter = history_filter(html_filter, 1000, 0.001)
        self.assertFalse(history_filter.accept("123231"))
        self.assertTrue(history_filter.accept("http://123123"))
        self.assertFalse(history_filter.accept("http://123123"))


if __name__ == '__main__':
    unittest.main()



import unittest
from test.tools import test_page
from aioresponses import CallbackResult
from easy_spider.network.request import Request
import re
from test.mock_env import env, get


class TestNetWork(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        env.mocked.get(re.compile(r"^http://localhost:5000/test_encoding.*$"),
                       content_type="text/html",
                       callback=TestNetWork.handler,
                       repeat=True)
        env.mocked.get("http://localhost:5000/test_request",
                       body='ok',
                       content_type="text/html",
                       repeat=True)

    @staticmethod
    def handler(url, **_):
        charset = url.query.get("charset", '')
        content_type = "text/html;charset={}".format(charset)
        return CallbackResult(
            body=test_page,
            content_type=content_type,
            status=200
        )

    def test_request(self):
        resp = get("http://localhost:5000/test_request")
        self.assertEqual(resp.text, "ok")

        resp = get("http://localhost:5000/test_request")
        self.assertEqual(resp.text, "ok")

    def test_set_encoding(self):
        text_response = get("http://localhost:5000/test_encoding")
        text_response.encoding = "utf-8"
        self.assertEqual(text_response.encoding, "utf-8")

    def test_infer_encoding(self):
        text_response = get("http://localhost:5000/test_encoding?charset=gb2312")
        self.assertEqual(text_response.encoding, 'gb18030')

        text_response = get("http://localhost:5000/test_encoding?charset=gbk")
        self.assertEqual(text_response.encoding, 'gb18030')

        text_response = get("http://localhost:5000/test_encoding?charset=utf-8")
        self.assertEqual(text_response.encoding, 'utf-8')

    def test_decode(self):
        # 测试编码错误的情况，此情况无法正常解码
        text_response = get("http://localhost:5000/test_encoding?charset=gb2312")
        self.assertNotIn("世界", text_response.text)

        # 测试编码正确的情况
        text_response = get("http://localhost:5000/test_encoding?charset=utf-8")
        self.assertIn("世界", text_response.text)

        # 测试编码不存在的情况
        text_response = get("http://localhost:5000/test_encoding")
        self.assertIn("世界", text_response.text)

        # 测试错误编码，但采用自动推断解码的情况
        text_response = get("http://localhost:5000/test_encoding?charset=gb2312")
        text_response.encoding = None  # 从内容自动推断 encoding
        self.assertIn("世界", text_response.text)

        # 测试给定不存在编码的情况
        text_response = get("http://localhost:5000/test_encoding?charset=ggg")
        self.assertIn("世界", text_response.text)
        text_response.encoding = "ggg"
        self.assertIn("世界", text_response.text)

    def test_cache(self):
        html_response = get("http://localhost:5000/test_encoding?charset=utf-8")
        self.assertIs(html_response.bs, html_response.bs)

    def test_ensure_request_callable(self):
        r = Request("test")
        with self.assertRaises(TypeError):
            r.handler = 1
        r.handler = lambda x: x

    def test_build_request(self):
        r = Request.of("http://test")
        self.assertEqual(r.uri, "http://test")
        r1 = Request.of(r)
        self.assertIs(r1, r)
        with self.assertRaises(TypeError):
            Request.of(1)


if __name__ == '__main__':
    unittest.main()



from easy_spider.filters.filter import *
from easy_spider.tool import get_extension
from bloom_filter import BloomFilter

# 非 html 后缀， 来源于 scrapy
# https://github.com/scrapy/scrapy/blob/master/scrapy/linkextractors/__init__.py
IGNORED_EXTENSIONS = {
    # archives
    '7z', '7zip', 'bz2', 'rar', 'tar', 'tar.gz', 'xz', 'zip',

    # images
    'mng', 'pct', 'bmp', 'gif', 'jpg', 'jpeg', 'png', 'pst', 'psp', 'tif',
    'tiff', 'ai', 'drw', 'dxf', 'eps', 'ps', 'svg', 'cdr', 'ico',

    # audio
    'mp3', 'wma', 'ogg', 'wav', 'ra', 'aac', 'mid', 'au', 'aiff',

    # video
    '3gp', 'asf', 'asx', 'avi', 'mov', 'mp4', 'mpg', 'qt', 'rm', 'swf', 'wmv',
    'm4a', 'm4v', 'flv', 'webm',

    # office suites
    'xls', 'xlsx', 'ppt', 'pptx', 'pps', 'doc', 'docx', 'odt', 'ods', 'odg',
    'odp',

    # other
    'css', 'pdf', 'exe', 'bin', 'rss', 'dmg', 'iso', 'apk', 'js'
}

static_filter = CustomFilter(lambda url: get_extension(url) in IGNORED_EXTENSIONS)
url_filter = RegexFilter(r"^https?:\/{2}[^\s]*?(\?.*)?$")
html_filter = url_filter - static_filter
all_pass_filter = CustomFilter(lambda _: True)
all_reject_filter = CustomFilter(lambda _: False)


class HistoryFilter(Filter):
    """
    布隆过滤器， 依赖于其他过滤器的结果
    pre_filter: 前置过滤器，布隆过滤器将依赖于前置过滤器的返回结果
    若布隆过滤器以及前置过滤器都返回 True 才记录到布隆过滤器中，并返回 True
    否则返回 False
    """
    def __init__(self, pre_filter: Filter, max_elements, error_rate):
        self._history_filter = BloomFilter(max_elements, error_rate)
        self._pre_filter = pre_filter or all_pass_filter

    def accept(self, url: str) -> bool:
        pre_filter_accept = self._pre_filter.accept(url)
        if not pre_filter_accept:  # 如果前置过滤器返回 False 则直接返回 False
            return False
        history_filter_accept = url not in self._history_filter
        history_filter_accept and self._history_filter.add(url)  # 如果不存在于布隆过滤器中，则记录
        return history_filter_accept


def history_filter(pre_filter: Filter, max_elements: int, error_rate: float):
    return HistoryFilter(pre_filter, max_elements, error_rate)



from easy_spider.core.spider import AsyncSpider
from easy_spider.log import logger
from easy_spider.network.request import RequestQueue, SimpleRequestQueue
import asyncio
from abc import ABC, abstractmethod


class Task(ABC):

    @abstractmethod
    def run(self): pass


class AbstractTask(Task, ABC):

    def __init__(self, spider, request_queue: RequestQueue):
        self._spider = spider
        self._request_queue = request_queue

    def _init_queue(self):
        if not self._spider.start_targets:
            raise ValueError("初始请求不能为空")
        self._request_queue.put_many(self._spider.start_targets)


# class MultiThreadJob(AbstractTask):
#     def __init__(self, spider: MultiThreadSpider, request_queue):
#         super().__init__(spider, request_queue)
#         self._thread_pool = ThreadPoolExecutor(max_workers=spider.start_requests)
#         self._num_running_task = Value("i", 0)
#         self._lock = Lock()
#
#     def run(self):
#         fs = [self._thread_pool.submit(self._run) for _ in range(self._spider.num_threads)]
#         wait(fs)
#
#     def _run(self):
#         while True:
#             with self._lock:
#                 resource = self._request_queue.get()
#                 if resource is None:
#                     if self._num_running_task.value == 0:  # 如果 task_queue 为空, 且同时正在进行的任务为0则退出
#                         break
#                     else:
#                         sleep(0)  # 否则 sleep 等待任务
#                         continue
#                 else:
#                     self._num_running_task.value += 1
#             try:
#                 new_resources = self._spider.crawl(resource)
#                 self._request_queue.put_many(new_resources)
#             except Exception as e:
#                 logger.warning(f"{resource}处理失败: {e}", exc_info=True)
#             finally:
#                 self._num_running_task.value -= 1


class AsyncTask(AbstractTask):
    def __init__(self, spider: AsyncSpider, request_queue=SimpleRequestQueue()):
        super().__init__(spider, request_queue)
        self._num_running_task = 0
        self._init_queue()

    async def run(self):
        await asyncio.gather(*[self._run() for _ in range(self._spider.num_threads)])

    async def _run(self):
        while True:
            request = self._request_queue.get()
            if request is None:
                if self._num_running_task != 0:
                    await asyncio.sleep(0)
                    continue
                else:  # 如果 task_queue 为空, 且同时正在进行的任务为 0 则退出
                    break
            try:
                self._num_running_task += 1
                new_requests = await self._spider.crawl(request)
                self._request_queue.put_many(new_requests)
                logger.info("成功 {} ".format(request))
            except Exception as e:
                logger.warning("失败 {}  原因 {}".format(request, e))
            finally:
                self._num_running_task -= 1



from easy_spider.core.env import async_env
from easy_spider.core.spider import AsyncSpider
from easy_spider.network.request import Request
from easy_spider.network.response import Response, HTMLResponse, TextResponse

__version__ = "1.0.3"
name = "easy_spider"



from abc import ABC, abstractmethod
from queue import Queue, Empty
from collections import deque


class Request(ABC):
    def __init__(self, uri, handler=None, method='GET', priority=0, tag="any",
                 headers=None, cookies=None, timeout=None, params=None, data=None,
                 data_format='form', encoding='utf-8'):
        self.uri: str = uri
        self._handler = handler
        self.priority: int = priority
        self.tag = tag
        self.method: str = method
        self.timeout: int = timeout
        self.headers: dict = headers or {}
        self.cookies: dict = cookies or {}
        self.encoding: str = encoding
        self.params: dict = params or {}
        self.data: dict = data
        self.data_format: str = data_format

    @property
    def handler(self):
        return self._handler

    @handler.setter
    def handler(self, callback):
        if not callable(callback):
            raise TypeError("handler must be a callable, but get type: `{}`".format(type(callback).__name__))
        self._handler = callback

    def __repr__(self):
        return "[method={} uri={}]".format(self.method, self.uri)

    def __str__(self):
        return self.__repr__()

    @staticmethod
    def of(instance):
        if isinstance(instance, Request):
            return instance
        elif isinstance(instance, str):
            return Request(instance)
        else:
            raise TypeError("can't build request from type `{}`".format(type(instance).__name__))


class RequestQueue(ABC):

    @abstractmethod
    def put(self, request: Request) -> None: pass

    def put_many(self, requests):
        for request in requests:
            self.put(request)

    @abstractmethod
    def get(self) -> Request: pass

    @abstractmethod
    def empty(self) -> bool: pass


class SyncRequestQueue(RequestQueue):

    def __init__(self):
        super().__init__()
        self._queue = Queue()

    def put(self, request: Request) -> None:
        self._queue.put(request)

    def get(self):
        try:
            return self._queue.get_nowait()
        except Empty:
            return None

    def empty(self) -> bool:
        return self._queue.empty()

    def __len__(self):
        return self._queue.qsize()


class SimpleRequestQueue(RequestQueue):

    def __init__(self):
        self._queue = deque()

    def put(self, request: Request) -> None:
        self._queue.append(request)

    def get(self) -> Request:
        try:
            return self._queue.pop()
        except IndexError:
            return None

    def empty(self) -> bool:
        return len(self) == 0

    def __len__(self):
        return len(self._queue)


import setuptools
from easy_spider import __version__

with open("README.md", encoding='utf-8') as fd:
    long_description = fd.read()

with open("requirements.txt", encoding="utf-8") as fd:
    install_requires = [line for line in fd.read().split("\n") if line]

packages = setuptools.find_packages(exclude=("test", "demo"))

setuptools.setup(
    name="easy-spider",
    version=__version__,
    author="lin3x",
    author_email="544670411@qq.com",
    description="A asynchronous spider with aiohttp",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://lin3x.coding.net/p/easy_spider",
    packages=packages,
    data_files=('readme.md', 'requirements.txt'),
    install_requires=install_requires,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from abc import ABC
from easy_spider.network.request import Request
import re
from typing import List


class Filter(ABC):

    def accept(self, url: str) -> bool: pass

    def __neg__(self):
        return NotFilter(self)

    def __add__(self, other):
        return AndChainFilter(self, other)

    def __or__(self, other):
        return OrChainFilter(self, other)

    def __sub__(self, other):
        return AndChainFilter(self, NotFilter(other))


class NotFilter(Filter):
    def __init__(self, filter):
        super().__init__()
        self.filter = filter

    def accept(self, url: str) -> bool:
        return not self.filter.accept(url)


class CustomFilter(Filter):
    def __init__(self, filter_func):
        super().__init__()
        self._filter_func = filter_func

    def accept(self, url: str) -> bool:
        return self._filter_func(url)


class RegexFilter(Filter):
    def __init__(self, re_expr):
        self._re_expr = re.compile(re_expr)

    def accept(self, url: str) -> bool:
        return bool(self._re_expr.match(url))


class AndChainFilter(Filter):
    def __init__(self, *filters: Filter):
        self._filters: List[Filter] = list(filters)

    def accept(self, url: str) -> bool:
        return all([f.accept(url) for f in self._filters])

    def __add__(self, other):
        self._filters.append(other)
        return self

    def __sub__(self, other):
        self._filters.append(NotFilter(other))
        return self


class OrChainFilter(Filter):
    def __init__(self, *filters: Filter):
        self._filters: List[Filter] = list(filters)

    def accept(self, url: str) -> bool:
        return any([f.accept(url) for f in self._filters])

    def __or__(self, other):
        self._filters.append(other)
        return self
